import datetime
import json

from ..renamer.printer import PrinterInterface
from ..utils.jsons import ExtendedJsonEncoder
from ..exceptions import CollectorError


# Non exhaustive list of Video containers with their file extension and name
# Should also describe some music only containers
MEDIAS_CONTAINERS = {
    "3gp": "3GPP",
    "asf": "Advanced Systems Format",
    "avi": "AVI",
    "flv": "Flash Video",
    "f4v": "Flash Video",
    "mov": "QuickTime",
    "mp4": "MPEG-4",
    "mkv": "Matroska",
    "mpg": "MPEG",
    "mpeg": "MPEG",
    "mpv": "MPEG",
    "mts": "MPEG Transport Stream",
    "qt": "QuickTime",
    "rm": "RealMedia",
    "ts": "MPEG Transport Stream",
    "vob": "Vob",
    "webm": "WebM",
    "wmv": "Windows Media Video",
}


# Default container label when extension does not match default container list
MEDIAS_DEFAULT_CONTAINER_NAME = "Unknow"


# List of unique file extensions for medias
MEDIAS_EXTENSIONS = set(MEDIAS_CONTAINERS.keys())


class Collector(PrinterInterface):
    """
    Collect informations about media files.

    Attributes:
        registry (dict): The registry where is collected all informations from scanning.
        stats (dict): Global statistics for all collected directories, files and total
            size.

    Arguments:
        basepath (pathlib.Path): The base directory for all directories to scan.
            Directories does not have to start directly from the basepath but must
            start from it.

    Keyword Arguments:
        extensions (list): A list of file extensions to consider as media files. A
            file extension that is not in this list are ignored. Default list contains
            all common video file extensions.
        allow_empty_dir (list): If True, even directory without direct media files are
            still collected in registry. Default is False, so only directories with
            direct media children are collected, the other directories are ignored from
            registry (but still scanned for children directories).
    """
    def __init__(self, basepath, extensions=MEDIAS_EXTENSIONS, allow_empty_dir=False):
        super().__init__()

        self.basepath = basepath
        self.extensions = extensions
        self.allow_empty_dir = allow_empty_dir
        self.reset()

    def reset(self):
        """
        Reset registry and global states.

        This is the method to use if you plan to make multiple usage of ``run`` or
        ``scan_directory`` for different basepath since registry and global states are
        cumulative.
        """
        self.registry = {}
        self.stats = {
            "directories": 0,
            "files": 0,
            "size": 0,
        }

    def store(self, data):
        """
        Store given directory data.

        Arguments:
            data (dict):
        """
        key = str(data["path"].relative_to(self.basepath))
        self.registry[key] = data

    def timestamp_to_isoformat(self, timestamp):
        """
        Return datetime formatted from given timestamp.

        Arguments:
            timestamp (float): A timestamp as expected from date returned in
                ``Path.stat()``.

        Returns:
            string: Datetime formatted in ISO format without microseconds.
        """
        return datetime.datetime.fromtimestamp(timestamp).isoformat(timespec="seconds")

    def scan_file(self, path):
        """
        Scan a media file to get its informations.

        Implemention does not support file without any suffix (file extension).

        Arguments:
            path (pathlib.Path): File path to scan for informations.

        Returns:
            dict: Collected file informations.
        """
        # Get file stats informations
        stats = path.stat()

        relative_dir = path.parent.relative_to(self.basepath)

        dirname = path.parent.name
        # Prefer empty dirname instead of basepath dirname when file is at basepath
        # root
        if dirname == self.basepath.name:
            dirname = ""

        # Remove leading dot
        extension = path.suffix[1:].lower()
        # Get the media container label from file extension
        container = MEDIAS_DEFAULT_CONTAINER_NAME
        if extension in MEDIAS_CONTAINERS:
            container = MEDIAS_CONTAINERS[extension]

        data = {
            "path": path,
            "name": path.name,
            "absolute_dir": path.parents[0],
            "relative_dir": relative_dir,
            "directory": dirname,
            "extension": extension,
            "container": container,
            "size": stats.st_size,
            "mtime": self.timestamp_to_isoformat(stats.st_mtime),
        }

        self.stats["files"] += 1
        self.stats["size"] += data["size"]

        return data

    def scan_directory(self, path):
        """
        Scan a directory to get its media files.

        Does not return anything, the directories informations (and possible media files
        informations) are stored in ``Collector.registry``.

        Arguments:
            path (pathlib.Path): Directory to scan for informations, for direct children
                files and to recursively search for children directories.

        Raises:
            CollectorError: If given path is not a directory inside
                basepath directory.
        """
        self.log_debug("Scanning {}".format(str(path)))

        try:
            relative_dir = path.relative_to(self.basepath)
        except ValueError:
            msg = "You cannot scan a directory which is out of given basepath: {}"
            raise CollectorError(msg.format(str(self.basepath)))

        # Get directory stats informations
        stats = path.stat()

        data = {
            "path": path,
            "name": path.name,
            "absolute_dir": path.parents[0],
            "relative_dir": relative_dir,
            "size": stats.st_size,
            "mtime": self.timestamp_to_isoformat(stats.st_mtime),
            "children_files": [],
        }

        for child in path.iterdir():
            if child.is_dir():
                self.scan_directory(child)
            else:
                if child.suffix and child.suffix.lower()[1:] in self.extensions:
                    data["children_files"].append(self.scan_file(child))

        if self.allow_empty_dir or len(data["children_files"]) > 0:
            self.stats["directories"] += 1
            self.stats["size"] += data["size"]
            self.store(data)

    def run(self, destination=None):
        """
        Recursively scan everything from basepath to produce a registry of collected
        informations.

        Keyword Arguments:
            destination (pathlib.Path): Destination path to write a JSON file with
                every collected informations. Default is ``None`` so no JSON registry
                file will be written to the filesystem.

        Returns:
            dict: Dictionnary of global states for collected directories and files.
        """
        self.scan_directory(self.basepath)

        if self.registry and destination:
            with destination.open("w") as fp:
                json.dump(self.registry, fp, indent=4, cls=ExtendedJsonEncoder)
                self.log_info("Registry saved to: {}".format(str(destination)))

        return self.stats



# -*- coding: utf-8 -*-
"""
Exceptions
==========

Specific application exceptions.
"""


class DeoviClientBaseException(Exception):
    """
    Exception base.

    You should never use it directly except for test purpose. Instead make or
    use a dedicated exception related to the error context.
    """
    pass


class DummyError(DeoviClientBaseException):
    """
    Dummy exception sample to raise from your code.
    """
    pass


class TaskValidationError(DeoviClientBaseException):
    """
    When task rules are not valid.
    """
    pass


class CollectorError(DeoviClientBaseException):
    """
    For any media file collector error.
    """
    pass


class JobValidationError(DeoviClientBaseException):
    """
    When a job is not valid.

    Keyword Arguments:
        validation_details (dict): A possible dictionnary of validation details. It
            won't output as exception message from traceback, you need to exploit it
            yourself if needed.
    """
    def __init__(self, *args, **kwargs):
        self.validation_details = kwargs.pop("validation_details", None)
        super().__init__(*args, **kwargs)



import logging


class PrinterInterface:
    """
    Printer interface to output messages.

    This is a common interface for objects to use logging and optional message
    formatting.
    """

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger("deovi")
        self.max_label = 0

    def set_label_length(self, tasks):
        """
        Crawl every task to get the bigger name length.

        Arguments:
            tasks (dict): A dictionnary of tasks infos, only its keys are used to
                compute max value.

        Returns:
            integer: Length of bigger name or zero if task list is empty.
        """
        if not tasks:
            return 0

        return max([len(k) for k, v in tasks])

    def get_indentation_infos(self, length, template):
        """
        Return computated indice and indentation informations.

        Arguments:
            length (integer): Length of items to compute maximum indentation value.
            template (string): String template to build formatted indice.

        Returns:
            tuple: In order, the indentation length (integer) and the indice formatter
                function. This one attempt only an integer value and return a string.
        """
        zfill = len(
            str(length)
        )

        indent_length = len(
            template.format(i=length)
        )

        indice_func = None

        # The formatter function carry the template to format indice directly from
        # given integer
        def indice_func(index):
            return template.format(i=str(index).zfill(zfill))

        return indent_length, indice_func

    def _row_format(self, message, label=None, state=None, indent=None):
        """
        Format row message from given arguments.

        Keyword Arguments:
            message (string): A message to display.
            state (string): State name can be start, default, debug or end. It will
                define how the row will be formatted.
            indent (integer or string): If a string, use it to prefix message. If an
                integer, define the number of space to be added as prefix.

        Returns:
            string: Formatted message.
        """
        prefix = ""

        if indent and isinstance(indent, int):
            prefix = " " * indent
        elif indent:
            prefix = indent

        # Select the leading string to insert before label or message
        start = "├─ "
        if state == "start":
            start = "┍━ "
        elif state == "debug":
            start = "├┄ "
        elif state == "end":
            start = "┕━ "

        # Add label surrounding and padding
        label = label or ""
        if label:
            template = "[{}]  "
            label = template.format(label)
            label = label.ljust(len(template.format("")) + self.max_label)

        # Put message if any
        message = message or ""

        return "".join([prefix, start, label, message])

    def log(self, level, message, **kwargs):
        """
        Log message at required level with or without formatting.

        If method receives either ``label`` or ``state`` keyword arguments, message
        will be formatted using ``_row_format`` (which keyword arguments are passed to)
        before to be logged.

        Else the message is just logged without formatting.

        Arguments:
            level (string): Logging level name (``debug``, ``info``, etc..).
        """
        if "label" in kwargs or "state" in kwargs or "indent" in kwargs:
            getattr(self.logger, level)(
                self._row_format(message, **kwargs)
            )
        else:
            getattr(self.logger, level)(message)

    def log_debug(self, *args, **kwargs):
        """
        Shortcut to log message at debug level.
        """
        self.log("debug", *args, **kwargs)

    def log_info(self, *args, **kwargs):
        """
        Shortcut to log message at info level.
        """
        self.log("info", *args, **kwargs)

    def log_warning(self, *args, **kwargs):
        """
        Shortcut to log message at warning level.
        """
        self.log("warning", *args, **kwargs)

    def log_error(self, *args, **kwargs):
        """
        Shortcut to log message at error level.
        """
        self.log("error", *args, **kwargs)

    def log_critical(self, *args, **kwargs):
        """
        Shortcut to log message at critical level.
        """
        self.log("critical", *args, **kwargs)



import datetime
import json
from pathlib import Path


class ExtendedJsonEncoder(json.JSONEncoder):
    """
    Additional opiniated support for more basic object types.
    """
    def default(self, obj):
        # Support for pathlib.Path to a string
        if isinstance(obj, Path):
            return str(obj)
        # Support for set to a list
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)



DUMMY_ISO_DATETIME = "1977-06-02T07:35:15"


def timestamp_to_isoformat(*args, **kwargs):
    """
    A dummy function which always returns the same string with datetime formatted to
    ISO format without microseconds.

    It is aimed to be used to mockup ``Collector.timestamp_to_isoformat`` in tests
    since sample files datetime are not stable through an installation to another.
    """
    return DUMMY_ISO_DATETIME



import logging
from pathlib import Path

import click

from ..collector import MEDIAS_EXTENSIONS, Collector


@click.command()
@click.argument("source", nargs=1, type=click.Path(exists=True, path_type=Path))
@click.argument("destination", nargs=1, type=click.Path(exists=False, path_type=Path))
@click.option("--extension", "-e", multiple=True)
@click.pass_context
def collect_command(context, source, destination, extension):
    """
    Recursively collect every media files from a basepath.
    """
    logger = logging.getLogger("deovi")

    if not extension:
        extension = MEDIAS_EXTENSIONS

    logger.info("Source: {}".format(source))
    logger.info("Destination: {}".format(destination))
    logger.info("Extensions: {}".format(", ".join(extension)))

    collector = Collector(source, extensions=extension)

    stats = collector.run(destination=destination)

    logger.info("Registered directories: {}".format(stats["directories"]))
    logger.info("Registered files: {}".format(stats["files"]))
    logger.info("Total directories and files size: {}".format(stats["size"]))



"""
Main entrance to commandline actions

TODO:
    R&D to implement "COMMAND -h" / "COMMAND --help" as a "help COMMAND" command.
"""
import click

from ..logger import init_logger
from ..cli.version import version_command
from ..cli.rename import rename_command
from ..cli.job import job_command
from ..cli.collect import collect_command


# Help alias on "-h" argument
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


# Default logger conf
APP_LOGGER_CONF = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", None)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v", "--verbose",
    type=click.IntRange(min=0, max=5),
    default=4,
    metavar="INTEGER",
    help=(
        "An integer between 0 and 5, where '0' make a totaly "
        "silent output and '5' set level to DEBUG (the most verbose "
        "level). Default to '4' (Info level)."
    )
)
@click.pass_context
def cli_frontend(ctx, verbose):
    """
    Entrypoint for commandline interface for deovi
    """
    printout = True
    if verbose == 0:
        verbose = 1
        printout = False

    # Verbosity is the inverse of logging levels
    levels = [item for item in APP_LOGGER_CONF]
    levels.reverse()
    # Init the logger config
    root_logger = init_logger(
        "deovi",
        levels[verbose],
        printout=printout
    )

    # Init the default context that will be passed to commands
    ctx.obj = {
        "verbosity": verbose,
        "logger": root_logger,
    }


# Attach commands methods to the main grouper
cli_frontend.add_command(version_command, name="version")
cli_frontend.add_command(rename_command, name="rename")
cli_frontend.add_command(job_command, name="job")
cli_frontend.add_command(collect_command, name="collect")



import json
from pathlib import Path

from ..exceptions import TaskValidationError


def validate_task_options(task_name, required, **options):
    """
    Validate that all options respect the task requirements.

    Arguments:
        task_name (string): Task name to output in possible error.
        required (list): List of required option names .
        **options (dict): Given task options.

    Raises:
        TaskValidationError: If required options are missing from given ``options``.
    """
    fields = []
    if required:
        for name in set(required):
            if name not in options:
                fields.append(name)

        if fields:
            msg = "Task '{name}' require option(s): {fields}".format(
                name=task_name,
                fields=", ".join(sorted(fields)),
            )
            raise TaskValidationError(msg)

    return True


def validate_job_file(filepath):
    """
    Validate if a Job files exist, is valid JSON and have required parameters.

    Arguments:
        filepath (pathlib.Path): Path to a file to check.

    Returns:
        list: Error messages if any.
    """
    errors = []

    # Don't continue checking if file does not exists
    if not filepath.exists():
        errors.append("Configuration file does not exists: {}".format(filepath))
    # Proceed to file checks
    else:
        # Don't continue checking if invalid JSON
        try:
            with filepath.open("r") as fp:
                data = json.load(fp)
        except json.decoder.JSONDecodeError as e:
            errors.append(
                "Configuration file is not a valid JSON file: {}\n{}".format(
                    filepath,
                    "   {}".format(str(e)),
                )
            )
        # Proceed to rules checking
        else:
            if "basepath" not in data:
                errors.append(
                    (
                        "Configuration file is missing required 'basepath' item: {}"
                    ).format(filepath)
                )
            # Validate if basepath exists and is a directory
            else:
                p = Path(data["basepath"])
                if not p.is_absolute():
                    p = (filepath.parents[0] / p).resolve()

                if not p.exists():
                    errors.append(
                        (
                            "Configuration file value for 'basepath' is not an "
                            "existing path: {}".format(p.resolve())
                        )
                    )
                elif not p.is_dir():
                    errors.append(
                        (
                            "Configuration file value for 'basepath' is not a "
                            "directory: {}".format(p)
                        )
                    )

            if "tasks" not in data:
                # NOTE: Task can still be an empty list, this is a behavior to permit
                #       to run a job just for listing basepath.
                errors.append(
                    (
                        "Configuration file miss required 'tasks' item: {}"
                    ).format(filepath)
                )

    return errors


def validate_jobs(filepaths):
    """
    Validate all jobs from given file paths.

    Arguments:
        filepaths (pathlib.Path): Path to a file to check.

    Returns:
        dict: All job errors where errors are indexed on their job files.
    """
    errors = {}

    for path in filepaths:
        results = validate_job_file(path)
        if results:
            errors[str(path)] = results

    return errors


def is_allowed_file(source, extensions=[]):
    """
    Check if source path is a file and allowed against a set of file extensions.

    A source file which does not exists on filesystem is not valid. Allowed extensions
    are compared only to the last file extension since some file names can contain dots
    as word divider but still recognized as extensions.

    Also, the comparaison is case insensitive.

    NOTE:
        Double extension like "tar.gz" won't work because of how this have been
        implemented. We match against extensions with the last suffix from Path object
        and so can only be "gz", this was done to be compatible with
        "dotted.file.names". But a new implementation may work correctly for double
        extension if using string method "endswith" instead of "Path.suffixes[-1]".

    Arguments:
        source (pathlib.Path): Path to a file.

    Keyword Arguments:
        extensions (list): A list of extensions to check against. If empty, every
            source filepaths are returned. Default to empty.
    """
    if not source.is_file():
        return False

    # If no extensions, no further check
    if not extensions:
        return True

    ext = source.suffixes[-1].lower()
    if ext and ext.startswith("."):
        ext = ext[1:]

    if ext not in extensions:
        return False

    return True



from .collect import (
    MEDIAS_CONTAINERS, MEDIAS_DEFAULT_CONTAINER_NAME, MEDIAS_EXTENSIONS, Collector,
)


__all__ = [
    "MEDIAS_CONTAINERS",
    "MEDIAS_DEFAULT_CONTAINER_NAME",
    "MEDIAS_EXTENSIONS",
    "Collector",
]



from ..exceptions import JobValidationError, TaskValidationError

from .printer import PrinterInterface
from .jobs import Job
from .validators import validate_jobs, validate_task_options


class JobRunner(PrinterInterface):
    """
    Validate then run Jobs.

    Arguments:
        task_class (tasks.TaskMaster): The task manager class to instanciate for each
            Job.
    """
    def __init__(self, task_class):
        super().__init__()

        self.task_class = task_class
        self.task_infos = None

    def get_available_tasks(self):
        """
        Get tasks informations (name and requirements) with inspection on task methods.

        Returns:
            dict: Dictionnary of tasks indexed on their name (the method name without
            suffix ``task_``) and with required attribute as value.
        """
        tasks = {}

        for item in dir(self.task_class):
            if item.startswith("task_"):
                name = item[len("task_"):]
                tasks[name] = []

                method = getattr(self.task_class, item)

                if hasattr(method, "required"):
                    tasks[name] = method.required

        return tasks

    def get_jobs(self, filepaths):
        """
        Load, inspect and validate every given Job filepaths.

        Validation is in two parts:

        1. Validate job parameters;
        2. Validate jobs tasks.

        The first one blocks the second one since invalid job parameters may not be
        suitable to open/parse for tasks.

        Arguments:
            filepaths (list): List of Path objects for job files.

        Raises:
            JobValidationError: If there is any error during job validation. The
                exception will contains the full detail of errors in
                ``validation_details`` attribute.

        Returns:
            list: List of Job model instances.
        """
        task_infos = self.get_available_tasks()
        job_errors = validate_jobs(filepaths)

        # If there is any error, stop execution with an exception resuming errors from
        # all invalid jobs
        if job_errors:
            raise JobValidationError(
                "Error(s) occured during job(s) validation",
                validation_details=job_errors,
            )

        # Load every job files in models
        jobs = [Job.load(item) for item in filepaths]

        # Validate tasks from all jobs
        task_by_job_errors = {}
        for job in jobs:
            job_source = str(job.source)

            # Check every job tasks
            for task_name, task_options in job.tasks:
                # First check job file exists else store error but dont continue to
                # task
                if task_name not in task_infos:
                    if job_source not in task_by_job_errors:
                        task_by_job_errors[job_source] = []
                    task_by_job_errors[job_source].append(
                        "There is no task with name '{}'.".format(task_name)
                    )
                    continue

                # Finally validate job tasks
                try:
                    validate_task_options(
                        task_name,
                        task_infos[task_name],
                        **task_options
                    )
                except TaskValidationError as e:
                    if job_source not in task_by_job_errors:
                        task_by_job_errors[job_source] = []
                    task_by_job_errors[job_source].append(str(e))

        # Raise exception with all task errors if any
        if task_by_job_errors:
            raise JobValidationError(
                "Error(s) occured during job(s) task(s) validation",
                validation_details=task_by_job_errors,
            )

        return jobs

    def run(self, filepaths, dry_run=True):
        """
        Run jobs from all given job filepaths.

        Arguments:
            filepaths (list): List of Path objects for job files.

        Keyword Arguments:
            dry_run (boolean): To enable "dry run" mode which does not write anything.
                If disabled, every renaming will be writed to the filesystem. Disabled
                by default.
        """
        task_manager = self.task_class()

        jobs = self.get_jobs(filepaths)

        # Perform all tasks from each jobs
        for job in jobs:
            job.run(task_manager, dry_run=dry_run)



import logging
from pathlib import Path

import click

from ..exceptions import JobValidationError
from ..renamer.jobs import Job


@click.command()
@click.argument("basepath", nargs=1, type=click.Path(exists=True, path_type=Path))
@click.option("--destination", type=click.Path(dir_okay=False, path_type=Path))
@click.pass_context
def job_command(context, basepath, destination):
    """
    Create a new empty Job file for given basepath (the directory where the Job will
    search files). If no destination is given, it will be the last basepath directory
    name slugified with ".json" extension.
    """
    logger = logging.getLogger("deovi")

    logger.debug("Basepath: {}".format(basepath))

    if destination:
        logger.debug("Destination: {}".format(destination))
    else:
        logger.debug("No destination given, automatic name from basepath will be used.")

    try:
        created = Job.create(basepath, destination=destination)
    except JobValidationError as e:
        logger.critical(e)
        raise click.Abort()
    else:
        logger.info("Created Job file: {}".format(created))



import json
from pathlib import Path

from slugify import slugify

from ..exceptions import JobValidationError

from .printer import PrinterInterface
from .validators import is_allowed_file


class Job(PrinterInterface):
    """
    Model object to store job parameters.

    Arguments:
        source (pathlib.Path): Job source filepath.
        basepath (pathlib.Path): Basepath for files to apply task.

    Keyword Arguments:
        name (string): Job name, if empty it will be set a dummy name with the source
            filepath.
        tasks (list): List of task to perform on files.
        extensions (list): List of file extensions to filter files. If empty, all file
            extensions are allowed.
        reverse (boolean): Enable reversion of file names when performing tasks.
    """
    DEFAULT_NAME = "Job from '{}'"
    EMPTY_JOB_MATRIX = {
        "name": None,
        "basepath": None,
        "extensions": None,
        "reversed": False,
        "tasks": [],
    }

    def __init__(self, source, basepath, name=None, tasks=None, extensions=None,
                 reverse=False):
        super().__init__()

        self.source = source
        self.name = name or self.DEFAULT_NAME.format(source)
        self.basepath = basepath
        self.tasks = tasks or []
        self.extensions = extensions
        self.reverse = reverse
        self.sort = True

    @classmethod
    def create(cls, basepath, destination=None):
        """
        Create a new empty Job file for given basepath.

        Arguments:
            basepath (string or pathlib.Path): Basepath to set in Job file. It does not
                matter if the basepath already exists or not. If given path is relative
                it will be resolved to an absolute path. Path must be an existing
                directory.

        Keyword Arguments:
            destination (string or pathlib.Path): Destination path for the Job file to
                create. Default to None, so the destination will be the slugified last
                basepath part (``foo`` from ``/home/foo``) at the current working
                directory. This is validated to not be an already existing path.

        Returns:
            pathlib.Path: Path for created Job file.
        """
        if not isinstance(basepath, Path):
            basepath = Path(basepath)

        # If not absolute, force basepath to be absolute from current working directory
        if not basepath.is_absolute():
            basepath = Path.cwd() / basepath
        basepath = basepath.resolve()

        if not basepath.is_dir():
            msg = "🚨 Basepath does not exists or is not a directory: {}"
            raise JobValidationError(msg.format(basepath))

        if not destination:
            # Craft filename from slugified basepath last part
            destination = Path(slugify(basepath.stem)).with_suffix(".json")
        elif not isinstance(destination, Path):
            destination = Path(destination)

        # If not absolute, force destination to be absolute from current working
        # directory
        if not destination.is_absolute():
            destination = Path.cwd() / destination
        destination = destination.resolve()

        # Finally check destination is a file and its directory exist
        if destination.is_dir():
            msg = "🚨 Destination path must be a file, not an existing directory: {}"
            raise JobValidationError(msg.format(destination))
        elif not destination.parents[0].exists():
            msg = "🚨 Destination path must point into an existing directory: {}"
            raise JobValidationError(msg.format(destination))

        content = Job.EMPTY_JOB_MATRIX.copy()
        content["basepath"] = str(basepath)

        with destination.open("w") as fp:
            json.dump(content, fp, indent=4)

        return destination

    @classmethod
    def load(cls, source):
        """
        Load Job parameters from a filepath and return a model instance.

        This assumes the job have already be validated before from
        ``validate_job_file`` validator.

        Arguments:
            source (string or pathlib.Path): Path to the Job file to load.

        Returns:
            Job: Instance of model configured with job parameters.
        """
        if not isinstance(source, Path):
            source = Path(source)

        with source.open("r") as fp:
            data = json.load(fp)

        # If relative path, resolve it from the source directory
        basepath = Path(data["basepath"])
        if not basepath.is_absolute():
            basepath = (source.parents[0] / basepath).resolve()

        return cls(
            source,
            basepath,
            name=data.get("name", None),
            tasks=data["tasks"],
            extensions=data.get("extensions", None),
            reverse=data.get("reversed", False),
        )

    def get_target_files(self):
        """
        Return list of eligible files from job basepath depending allowed extensions.

        This is not recursive.

        Returns:
            list: List of Path objects eligible files.
        """
        files = []

        msg = self.extensions or ["All"]
        self.log_info(" • Allowed file extension(s): {}".format(", ".join(msg)))

        for filename in self.basepath.iterdir():
            if is_allowed_file(filename, extensions=self.extensions):
                files.append(filename)

        if self.sort:
            files = sorted(files)

        if self.reverse:
            files = reversed(files)

        self.log_info(" • {} files to process".format(len(files)))

        return files

    def run(self, task_manager, dry_run=True):
        """
        Run all job tasks.

        Arguments:
            task_manager (tasks.TaskMaster): The task manager to use to run tasks.

        Keyword Arguments:
            dry_run (boolean): To enable "dry run" mode which does not write anything.
                If disabled, every renaming will be writed to the filesystem. Disabled
                by default.

        Returns:
            dict: A dictionnary of informations about original sources and renaming.
        """
        # A list of renamed filenames used for statistic and warning about overwritting
        # in dry run mode
        rename_store = set([])
        # A list of tuple "(renamed, original)" to restore everything if needed
        reverse_store = []

        self.log_info("📂 Working on: {}".format(self.basepath))

        # A list of original source filenames
        original_store = self.get_target_files()
        self.log_info("")

        # Build row indice and indentation stuff. Indentation length is computed
        # from indice string for the biggest index value
        indentation, indicer = self.get_indentation_infos(
            len(original_store), "[{i}]━",
        )

        # Safely run task, in case of any exception, renamed file will be restored to
        # their original name
        try:
            for i, source in enumerate(original_store, start=1):
                msg = "From: {}".format(source.name)
                self.log_info(msg, state="start", indent=indicer(i))
                # Start destination from original which will be update along the tasks
                destination = source

                # Perform task to alter destination to final renamed filename
                for name, options in self.tasks:
                    task_method = "task_{}".format(name)

                    # Perform task
                    paths = getattr(task_manager, task_method)(
                        i,  destination, _indent=indentation, **options
                    )

                    # Update destination filename from returned task value
                    destination = paths[1]

                # If final destination is not the same as the original
                if destination != source:

                    # Error if file already exists, no overwritting is allowed
                    if destination.exists():
                        msg = (
                            "❗ Destination already exists and won't be overwritten: {}"
                        ).format(
                            destination.name
                        )
                        self.log_warning(msg, state="end", indent=indentation)
                    elif destination in rename_store:
                        msg = (
                            "❗ This destination is already planned from another file:"
                            " {}"
                        ).format(
                            destination.name
                        )
                        self.log_warning(msg, state="end", indent=indentation)
                    # Dry run mode just output results without writing anything
                    elif dry_run:
                        msg = "✨ To: {}".format(destination.name)
                        self.log_info(msg, state="end", indent=indentation)
                        rename_store.add(destination)
                    # Effective renaming
                    else:
                        msg = "✅ To: {}".format(destination.name)
                        self.log_info(msg, state="end", indent=indentation)
                        reverse_store.append(
                            (source, destination)
                        )
                        # Perform renaming
                        source.rename(destination)
                        rename_store.add(destination)
                    # White space divider between jobs
                    self.log_info("")
                # If source and destination are identical, nothing to do, just warning
                else:
                    msg = (
                        "❗ Source and destination paths are identical, nothing to "
                        "rename."
                    )
                    self.log_warning(msg, state="end", indent=indentation)
                    self.log_warning("")
        except:  # noqa: E722
            # Restore every renamed files before error
            msg = (
                "🚨 An error occured during a job, every files will be "
                "restored with their original filename."
            )
            self.log_error("")
            self.log_error(msg)
            self.log_error("")

            # Restore renamed file to their original filenames
            for source, destination in reverse_store:
                self.log_error("*", destination, "=>", source)
                source.rename(destination, source)
            self.log_error("")

            # Finally raise the original error
            raise

        return {
            "original_store": original_store,
            "rename_store": rename_store,
        }



from .printer import PrinterInterface


class TaskMethods:
    """
    Task method container mixin to inherit for basic tasks.

    All method which start with "task_" is assumed to be a task method. Task
    methods should set an attribute "required" for validation, if this attribute is not
    set or empty, we assume there is no validation to do for the task.

    Note than task NEVER write anything, they just returns a tuple with original source
    filename as first item and second item is the resulting filename from task.

    All tasks assume given path are files, no further validation will be done on given
    file paths.

    Also tasks only works with Path objects, no string for file paths are allowed.

    Keyword Arguments:
        formatter (object):
    """

    def task_capitalize(self, index, source, **options):
        """
        Capitalize file name.

        Work on the whole filename including extension.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.

            source (pathlib.Path): Source filename.
            **options (dict): Tasks options, this task does not expect any options.

        Returns:
            tuple: A tuple of Path objects for respectively original and renamed
            filename Path.
        """
        destination = source.with_name(source.name.capitalize())

        self.log_debug(
            destination.name,
            label="capitalize",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)

    def task_lowercase(self, index, source, **options):
        """
        Lowercase file name.

        Work on the whole filename including extension.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.

            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, this task does not expect any options.

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        destination = source.with_name(source.name.lower())

        self.log_debug(
            destination.name,
            label="lowercase",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)

    def task_uppercase(self, index, source, **options):
        """
        Uppercase file name.

        Work on the whole filename including extension.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.

            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, this task does not expect any options.

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        destination = source.with_name(source.name.upper())

        self.log_debug(
            destination.name,
            label="uppercase",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)

    def task_underscore_to_dash(self, index, source, **options):
        """
        Convert some strings into another ones:

        * "_" to "-";
        * "---" to "_";

        So "ping_-_foo_bar.mp4" will be converted to "ping_foo-bar.mp4".

        Work on the whole filename including extension.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.

            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, this task does not expect any options.

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        destination = source.with_name(
            source.name.replace("_", "-").replace("---", "_")
        )

        self.log_debug(
            destination.name,
            label="underscore_to_dash",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)

    def task_add_prefix(self, index, source, **options):
        """
        Add a prefix before filename.

        Work on the whole filename including extension.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.

            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, expect an option "prefix".

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        prefix = options["prefix"]

        destination = source.with_name(prefix + source.name)

        self.log_debug(
            destination.name,
            label="add_prefix",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)
    task_add_prefix.required = ["prefix"]

    def task_numerate(self, index, source, **options):
        """
        Add current index position of item in file list.

        Index number is filled (from given zfill length) to the right with '0' and
        divided from filename with given divider string.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.

            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, expect option "zfill" and optional
                "start" and "divider".

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        zfill = options["zfill"]
        start = options.get("start", 0)
        divider = options.get("divider", "_")

        new_filename = "{index}{divider}{filename}".format(
            index=str(start + index).zfill(zfill),
            divider=divider,
            filename=source.name,
        )

        destination = source.with_name(new_filename)

        self.log_debug(
            destination.name,
            label="numerate",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)
    task_numerate.required = ["zfill"]

    def task_catch_segments(self, index, source, **options):
        """
        Catch segments from filename.

        Filename is splitted on a divider and only required segments from given rules
        are retained then joined.

        This only apply on the file name part, not the extension parts.

        Joiner option  can use a custom string to join segments, default string if not
        given is "-".

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
                used for print out message.
            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, require options "divider" and
                "slice_start" and accept optional "slice_end" and "joiner".

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        divider = options["divider"]
        joiner = options.get("joiner", "-")
        slice_start = options["slice_start"]
        slice_end = options.get("slice_end", None)

        # Get the filename root, eg: without file extension
        root = source.stem
        # Get the file extension if any
        ext = ""
        if len(source.suffixes) > 0:
            ext = source.suffixes[-1]

        # Split name on divider
        segments = root.split(divider)
        # Only remove segment if there is more than one
        if len(segments) > 1:
            if slice_end:
                segments = segments[slice_start:slice_end]
            else:
                segments = segments[slice_start:]

        # Add again the file extension to modified filename root
        new_filename = joiner.join(segments) + ext

        destination = source.with_name(new_filename)

        self.log_debug(
            destination.name,
            label="catch_segments",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)
    task_catch_segments.required = ["divider", "slice_start"]

    def task_replace(self, index, source, **options):
        """
        Replace all occurences in a string (from) by another one (to).

        This is applied on the whole filename, including the file extensions.

        Arguments:
            index (integer): Index integer for current item in the walked list. Mostly
            used for print out message.
            source (pathlib.Path): Source Path object.
            **options (dict): Tasks options, expect options "from" and "to".

        Keyword Arguments:
            verbose (string): To enable or disable verbosity, when disable not any
                output will be printed out. Default to False.
            printer (function): A function to use to print out message. If not set,
                ``verbose`` argument is useless. Default to None.

        Returns:
            tuple: A tuple of current filename and new filename.
        """
        from_string = options["from"]
        to_string = options["to"]

        destination = source.with_name(
            source.name.replace(from_string, to_string)
        )

        self.log_debug(
            destination.name,
            label="replace",
            state="debug",
            indent=options.get("_indent"),
        )

        return (source, destination)
    task_replace.required = ["from", "to"]


class TaskMaster(PrinterInterface, TaskMethods):
    """
    Task manager glues classes for printer and method implementations.
    """
    pass



import logging
from pathlib import Path

import click

from ..renamer.tasks import TaskMaster
from ..renamer.runner import JobRunner
from ..exceptions import JobValidationError


@click.command()
@click.argument("jobs", nargs=-1, type=click.Path(exists=True, path_type=Path))
@click.option('--commit', is_flag=True)
@click.pass_context
def rename_command(context, jobs, commit):
    """
    Rename multiple files with some tasks in a job file.

    A job file is a JSON file with some parameters and tasks to define formatting. Job
    does not recursively walk on files from its basepath directory.

    Default behavior is to not change anything, just list task it would do and
    you need to enable commit with the '--commit' argument.

    You can't mix job and init modes and so mix Job filepath and directory
    paths.

    Finally, if any job task fails all renamed file will be reversed to their
    original filename.
    """
    logger = logging.getLogger("deovi")

    if not jobs:
        logger.critical("There is no job file to process.")
        raise click.Abort()
    else:
        if commit:
            logger.info("Commit mode is enabled, file will be renamed.")
        else:
            logger.info("Dry run mode is enabled, no file will be renamed.")
        logger.info("")

        jobber = JobRunner(task_class=TaskMaster)

        try:
            jobs = jobber.run(jobs, dry_run=not(commit))
        except JobValidationError as e:
            logger.error("🚨 {}".format(str(e)))
            logger.error("")

            for jobpath, errors in e.validation_details.items():
                logger.error("📂 {}".format(jobpath))
                for item in errors:
                    logger.error(" • {}".format(item))
                logger.error("")

            raise click.Abort()
